from abc import ABC, abstractmethod
from dataclasses import dataclass

from ..constants import CALENDAR_PREFIX, DEVICE_PREFIX, DRIVER_PREFIX
from ..type_helpers import Get, ResponseSet, Set


class API(ABC):
    @abstractmethod
    def create_get_request(self) -> Get:
        raise NotImplementedError()

    @abstractmethod
    def parse(self, responses: ResponseSet):
        raise NotImplementedError()


@dataclass
class _APIBase(API):
    name: str
    access: str
    param: bool

    @property
    def _url(self) -> str:
        raise NotImplementedError()

    def create_get_request(self) -> Get:
        return Get(path=self._url)

    def create_set_request(self, value: str) -> Set:
        return Set(path=self._url, value=str(value))

    def parse(self, responses: ResponseSet) -> str | None:
        r = responses.get(self._url)
        return r.value if r else None


@dataclass
class _APIBaseExt(_APIBase):
    typ: str
    structure_id: int
    offset: int
    mask: int | None


@dataclass
class PageAPI(_APIBase):
    structure_id: int


@dataclass
class ScenarioAPI(_APIBaseExt):
    pass


@dataclass
class DriverAPI(_APIBaseExt):
    history: str | None

    @property
    def _url(self) -> str:
        if self.mask:
            return f"{DRIVER_PREFIX}/{self.structure_id}/{self.offset}/{self.mask}"
        else:
            return f"{DRIVER_PREFIX}/{self.structure_id}/{self.offset}"


@dataclass
class CalendarAPI(_APIBaseExt):
    calendar_type: str | None

    @property
    def _url(self) -> str:
        if self.mask:
            return f"{CALENDAR_PREFIX}/{self.structure_id}/{self.offset}/{self.mask}"
        else:
            return f"{CALENDAR_PREFIX}/{self.structure_id}/{self.offset}"


@dataclass
class DeviceAPI(_APIBase):
    typ: str
    device_id: int
    device_structure_id: int
    offset: int
    mask: int | None

    @property
    def _url(self) -> str:
        if self.mask:
            return f"{DEVICE_PREFIX}/{self.device_structure_id}/{self.offset}/{self.mask}"
        else:
            return f"{DEVICE_PREFIX}/{self.device_structure_id}/{self.offset}"



import contextlib
import re
from dataclasses import dataclass, fields
from xml.etree import ElementTree

import requests

from .api.generic import API, CalendarAPI, DeviceAPI, DriverAPI, PageAPI, ScenarioAPI
from .api.room import RoomAPI, RoomState
from .api.sunblind import SunblindAPI, SunblindState
from .constants import API_PATH, MAX_RESPONSE_LENGTH, REGEXP, SYSTEM_VARS, XML_PATH
from .type_helpers import Get, RequestSet, Response, ResponseSet


def parse_responses(r: str) -> ResponseSet:
    lines = r.splitlines()
    responses = [Response(*line.split("=")) for line in lines]
    response_set = {}
    for r in responses:
        response_set[r.path] = r
    return response_set


@dataclass
class SystemState:
    failure: bool
    system_ok: bool
    out_temperature: float
    latitude: float
    longitude: float
    system_time: int
    set_heat: bool
    system_on: bool
    drivers_ok: bool
    devices_ok: bool
    all_ok: bool

    def __post_init__(self):
        # Try to coerce types
        for field in fields(self):
            if callable(field.type):
                with contextlib.suppress(ValueError):
                    if field.type is bool:
                        setattr(self, field.name, field.type(int(getattr(self, field.name))))
                    else:
                        setattr(self, field.name, field.type(getattr(self, field.name)))


@dataclass
class State:
    system: SystemState
    rooms: dict[str, RoomState]
    sunblinds: dict[str, SunblindState]


class Controller:
    name: str
    rooms: dict[str, RoomAPI]
    sunblinds: dict[str, SunblindAPI]

    _url: str
    _room_ids: list[str]
    _sunblind_ids: list[str]
    _drivers_by_name: dict[str, DriverAPI]
    _devices_by_name: dict[str, DeviceAPI]
    _pages_by_name: dict[str, PageAPI]
    _system_drivers: dict[str, DriverAPI]

    def __init__(self, host: str, name: str = "IQtec Controller", proto: str = "http"):
        self._url = f"{proto}://{host}"
        self.name = name

        # Connect to the Device and set up apis
        apis = self._get_apis()
        self._drivers_by_name = {}
        self._devices_by_name = {}
        self._pages_by_name = {}
        for api in apis:
            match api:
                case DriverAPI():
                    self._drivers_by_name[api.name] = api
                case DeviceAPI():
                    self._devices_by_name[api.name] = api
                case PageAPI():
                    self._pages_by_name[api.name] = api
        # Setup System
        self._system_drivers = {}
        for var in SYSTEM_VARS:
            d = self._drivers_by_name.get(f"SYSTEM.{var.value}")
            if d:
                self._system_drivers[var.name] = d
        # Setup Rooms
        self._find_rooms()
        self._create_rooms()
        # Setup Sunblinds
        self._find_sunblinds()
        self._create_sunblinds()

    def _get_xml(self) -> ElementTree.Element:
        response = requests.get(self._url + XML_PATH)
        return ElementTree.fromstring(response.content)

    def _get_apis(self) -> list[API]:
        xml = self._get_xml()
        drivers = []
        for child in xml:
            match child.attrib:
                case {"category": "driver", **rest}:
                    m = rest.get("mask", None)
                    drivers.append(
                        DriverAPI(
                            name=str(rest.get("name")),
                            access=str(rest.get("access")),
                            param=bool(int(rest.get("param"))),
                            typ=str(rest.get("type")),
                            structure_id=int(rest.get("structure_id")),
                            offset=int(rest.get("offset")),
                            mask=int(m) if m else None,
                            history=rest.get("history", None),
                        )
                    )
                case {"category": "calendar", **rest}:
                    m = rest.get("mask", None)
                    drivers.append(
                        CalendarAPI(
                            name=str(rest.get("name")),
                            access=str(rest.get("access")),
                            param=bool(int(rest.get("param"))),
                            typ=str(rest.get("type")),
                            structure_id=int(rest.get("structure_id")),
                            offset=int(rest.get("offset")),
                            mask=int(m) if m else None,
                            calendar_type=rest.get("history", None),
                        )
                    )
                case {"category": "device", **rest}:
                    m = rest.get("mask", None)
                    drivers.append(
                        DeviceAPI(
                            name=str(rest.get("name")),
                            access=str(rest.get("access")),
                            param=bool(int(rest.get("param"))),
                            typ=str(rest.get("type")),
                            device_id=int(rest.get("device_id")),
                            device_structure_id=int(rest.get("device_structure_id")),
                            offset=int(rest.get("offset")),
                            mask=int(m) if m else None,
                        )
                    )
                case {"category": "sbScenario", **rest}:
                    m = rest.get("mask", None)
                    drivers.append(
                        ScenarioAPI(
                            name=str(rest.get("name")),
                            access=str(rest.get("access")),
                            param=bool(int(rest.get("param"))),
                            typ=str(rest.get("type")),
                            structure_id=int(rest.get("structure_id")),
                            offset=int(rest.get("offset")),
                            mask=int(m) if m else None,
                        )
                    )
                case {"category": "page", **rest}:
                    drivers.append(
                        PageAPI(
                            name=str(rest.get("name")),
                            access=str(rest.get("access")),
                            param=bool(int(rest.get("param"))),
                            structure_id=int(rest.get("structure_id")),
                        )
                    )
                case _:
                    raise NotImplementedError(f"Cannot parse driver {child.attrib}")
        return drivers

    def _find_rooms(self):
        rooms = set()
        self._room_ids = []
        for n in self._drivers_by_name:
            prefix = n.split(".")[0]
            if re.match(REGEXP.ROOM, prefix):
                rooms.add(prefix)
        self._room_ids = sorted(rooms)

    def _create_rooms(self):
        self.rooms = {}
        for room_id in self._room_ids:
            room = RoomAPI(room_id, self._drivers_by_name)
            self.rooms[room_id] = room

    def _find_sunblinds(self):
        blinds = set()
        self._sunblind_ids = []
        for n in self._drivers_by_name:
            prefix = n.split(".")[0]
            if re.match(REGEXP.SUNBLIND, prefix):
                blinds.add(prefix)
        self._sunblind_ids = sorted(blinds)

    def _create_sunblinds(self):
        self.sunblinds = {}
        for sunblind_id in self._sunblind_ids:
            sunblind = SunblindAPI(sunblind_id, self._drivers_by_name)
            self.sunblinds[sunblind_id] = sunblind

    def create_system_requests(self) -> list[Get]:
        return [self._system_drivers[f.name].create_get_request() for f in fields(SystemState)]

    def parse_system(self, response_set: ResponseSet) -> SystemState:
        return SystemState(**{f.name: self._system_drivers[f.name].parse(response_set) for f in fields(SystemState)})

    def api_call(self, request_set: RequestSet) -> ResponseSet:
        # Split GET to chunks
        chunks = []
        current_chunk = []
        current_chunk_length = 0
        for getter in request_set.getters:
            expected_length = getter.expected_length if getter.expected_length else 1
            if expected_length > MAX_RESPONSE_LENGTH:
                raise ValueError(
                    f"Expected response for {getter.path} is too long ({getter.expected_length}>{MAX_RESPONSE_LENGTH=})"
                )
            if current_chunk_length + expected_length > MAX_RESPONSE_LENGTH:
                path_string = ";".join(g.path for g in current_chunk)
                chunks.append(path_string)
                current_chunk = [getter]
                current_chunk_length = expected_length
            else:
                current_chunk.append(getter)
                current_chunk_length += expected_length
        if current_chunk:
            path_string = ";".join(g.path for g in current_chunk)
            chunks.append(path_string)

        path_set = ";".join([f"{r.path}={r.value}" for r in request_set.setters])

        responses = {}
        for chunk in chunks:
            url = f"{self._url}{API_PATH}{chunk}"
            r = requests.get(url)
            responses.update(parse_responses(r.text))

        if path_set:
            url = f"{self._url}{API_PATH}{path_set}"
            r = requests.get(url)
            responses.update(parse_responses(r.text))

        return responses

    def update(self, api: API):
        get = api.create_get_request()
        request = RequestSet([get], [])
        response = self.api_call(request)
        return api.parse(response)

    def update_system(self) -> SystemState:
        get_system = self.create_system_requests()
        request = RequestSet(get_system, [])
        response = self.api_call(request)
        return self.parse_system(response)

    def update_status(self) -> State:
        get_system = self.create_system_requests()
        get_rooms = [r.create_get_request() for r in self.rooms.values()]
        get_sunblinds = [r.create_get_request() for r in self.sunblinds.values()]
        request = RequestSet(get_system + get_rooms + get_sunblinds, [])
        response = self.api_call(request)
        return State(
            system=self.parse_system(response),
            rooms={r_id: r.parse(response) for r_id, r in self.rooms.items()},
            sunblinds={s_id: s.parse(response) for s_id, s in self.sunblinds.items()},
        )



import contextlib
from dataclasses import dataclass, fields

from ..constants import DRIVER_PREFIX, ROOM_VARS
from ..type_helpers import Get, ResponseSet
from .generic import API, DriverAPI


@dataclass
class RoomState:
    fan_command: int
    name: str
    eco_mode: bool
    holiday: bool
    important: bool
    calendar_number: int
    room_mode: int
    corr_time: int
    winter_holiday: float
    summer_holiday: float
    min_humidity: float
    max_humidity: float
    actual_temperature: float
    requested_temperature: float | str
    failure: bool
    humidity_problem: bool
    open_window: bool
    open_window_time: int
    open_window_midnight: bool
    at_home: bool
    weekend: bool
    key: bool
    humidity_low: bool
    humidity_high: bool
    heating: bool
    cooling: bool
    cooling_enabled: bool
    heating_enabled: bool
    cooling_mode: bool
    manual_correction: bool
    light_on: bool
    at_least_one_up: bool
    at_least_one_down: bool
    humidity: float
    first_symbol: int
    last_symbol: int
    heating_type: int
    correction_status: int
    correction_time: int
    correction_temperature: float
    calendar_temperature: float

    def __post_init__(self):
        # Try to coerce types
        for field in fields(self):
            if callable(field.type):
                with contextlib.suppress(ValueError):
                    if field.type is bool:
                        setattr(self, field.name, field.type(int(getattr(self, field.name))))
                    else:
                        setattr(self, field.name, field.type(getattr(self, field.name)))

        with contextlib.suppress(ValueError):
            # handle self.requested_temperature separately
            self.requested_temperature = float(self.requested_temperature)


class RoomAPI(API):
    room_id: str

    _drivers: dict[str, DriverAPI]
    _room_url: str

    def __init__(self, room_id: str, drivers: dict[str, DriverAPI]):
        self.room_id = room_id
        # Map room drivers
        self._drivers = {}
        for var in ROOM_VARS:
            d = drivers.get(f"{room_id}.{var.value}")
            if d:
                self._drivers[var.name] = d
                # # Dynamically create methods
                # setattr(self, f"get_{var.name}_request", self._drivers[var.name].create_get_request)
                # setattr(self, f"set_{var.name}_request", self._drivers[var.name].create_set_request)
                # setattr(self, f"parse_{var.name}t", self._drivers[var.name].parse)

        self._room_url = f"{DRIVER_PREFIX}/{self._drivers['name'].structure_id}"

    def create_get_request(self) -> Get:
        return Get(path=self._room_url, expected_length=len(self._drivers))

    def parse(self, response_set: ResponseSet):
        return RoomState(**{key: driver.parse(response_set) for key, driver in self._drivers.items()})



from dataclasses import dataclass

type Request = Get | Set

type ResponseSet = dict[str, Response]


@dataclass
class Response:
    path: str
    value: str


@dataclass
class Get:
    path: str
    expected_length: int | None = None


@dataclass
class Set:
    path: str
    value: str


@dataclass
class RequestSet:
    getters: list[Get]
    setters: list[Set]



import contextlib
from dataclasses import dataclass, fields

from ..constants import DRIVER_PREFIX, SUNBLIND_VARS
from ..type_helpers import Get, ResponseSet
from .generic import API, DriverAPI


@dataclass
class SunblindState:
    failure: bool
    out_up_1: bool
    out_up_2: bool
    out_dn_1: bool
    out_dn_2: bool
    manual_up: bool
    manual_dn: bool
    en: bool
    dis: bool
    urgent_up: bool
    sun_dn: bool
    dn: bool
    up: bool
    step_dn: bool
    step_up: bool
    calendar_en: bool
    move_time: int
    reverse_time: int
    tilt_time: int
    short_down_time: int
    step_time: int
    full_time_time: int
    name: str
    dead_time: int
    state: int
    rotation: int
    position: int
    command: int
    state2: int

    def __post_init__(self):
        # Try to coerce types
        for field in fields(self):
            if callable(field.type):
                with contextlib.suppress(ValueError):
                    if field.type is bool:
                        setattr(self, field.name, field.type(int(getattr(self, field.name))))
                    else:
                        setattr(self, field.name, field.type(getattr(self, field.name)))


class SunblindAPI(API):
    sunblind_id: str

    _drivers: dict[str, DriverAPI]
    _sunblind_url: str

    def __init__(self, sunblind_id: str, drivers: dict[str, DriverAPI]):
        self.sunblind_id = sunblind_id
        # Map room drivers
        self._drivers = {}
        for var in SUNBLIND_VARS:
            d = drivers.get(f"{sunblind_id}.{var.value}")
            if d:
                self._drivers[var.name] = d

        self._sunblind_url = f"{DRIVER_PREFIX}/{self._drivers['name'].structure_id}"

    def create_get_request(self) -> Get:
        return Get(path=self._sunblind_url, expected_length=len(self._drivers))

    def parse(self, response_set: ResponseSet):
        return SunblindState(**{key: driver.parse(response_set) for key, driver in self._drivers.items()})



from enum import IntEnum, StrEnum

MAX_RESPONSE_LENGTH = 100

API_PATH = "/control/?"

XML_PATH = "/proj/data.xml"

DRIVER_PREFIX = "1"

CALENDAR_PREFIX = "2"

DEVICE_PREFIX = "5"


class REGEXP:
    ROOM = r"^R\d{1,2}$"
    SUNBLIND = r"\w{0,}\_SUNBLIND\w{0,}"


class SYSTEM_VARS(StrEnum):
    failure = "_Failure"
    relay_check = "RelayCheck"
    hdo = "HDO"
    holiday_on = "HolidayON"
    holiday_off = "HolidayOFF"
    holiday_on_confirm = "HolidayONconfirm"
    holiday_off_confirm = "HolidayOFFconfirm"
    system_ok = "SystemOK"
    out_temperature = "OutTempearture"  # Note: Typo in the original XML source
    fltr = "Filter"
    accuracy = "Accuracy"
    no_move_time = "NoMoveTime"
    moving_time = "MovingTime"
    link_t1_can = "LinkT1_CAN"
    link_t2_can = "LinkT2_CAN"
    user1 = "USER1"
    user2 = "USER2"
    user3 = "USER3"
    user4 = "USER4"
    link_t3_can = "LinkT3_CAN"
    open_all_valves = "OpenAllValves"
    latitude = "Latitude"
    longitude = "Longtitude"  # Note: Typo in the original XML source
    system_time = "SystemTime"
    system_time_no_sec = "SystemTimeNoSec"
    departure_time = "DepartureTime"
    departure_active = "DepartureActive"
    arrival_time = "ArrivalTime"
    arrival_active = "ArrivalActive"
    holiday_active = "HolidayActive"
    set_heat = "SET_HEAT"
    set_cool = "SET_COOL"
    set_trunk = "SET_TRUNK"
    trunk_active = "TRUNK_ACTIVE"
    set_user1 = "SET_USER1"
    set_user2 = "SET_USER2"
    set_user3 = "SET_USER3"
    set_user4 = "SET_USER4"
    heat_on = "HeatOn"
    cool_on = "CoolOn"
    on_battery = "OnBattery"
    battery_failure = "BatteryFailure"
    system_on = "SystemOn"
    drivers_ok = "DriversOk"
    devices_ok = "DevicesOk"
    all_ok = "AllOk"
    time_minutes = "TIME_MINUTES"
    time_minutes_ws = "TIME_MINUTES_WS"
    sunrise_minutes = "SUNRISE_MINUTES"
    sunset_minutes = "SUNSET_MINUTES"
    set_person_home = "SET_PERSON_HOME"
    set_coming_home = "SET_COMMING_HOME"  # Note: Typo in original XML source
    set_room_party = "SET_ROOM_PARTY"
    person1_home = "Person1home"
    person2_home = "Person2home"
    person3_home = "Person3home"
    person4_home = "Person4home"
    person5_home = "Person5home"
    person6_home = "Person6home"
    person7_home = "Person7home"
    person8_home = "Person8home"
    open_window = "OpenWindow"
    light_on = "LightOn"
    at_least_one_up = "AtLeastOneUp"
    at_least_one_down = "AtLeastOneDown"
    sw_version = "SWversion"
    sw_subversion = "SWsubversion"
    booter_sw_version = "BooterSWversion"
    cpu_temperature = "CPUtemperature"
    dw_temperature = "DWtemperature"
    vbat = "VBAT"
    u24 = "U24"
    crc32 = "_CRC32"
    crc32_calendars = "_CRC32_Calendars"
    system_command = "SystemCommand"
    loop_time_us = "LoopTime_us"
    fast_drivers_us = "FastDrivers_us"
    slow_drivers_us = "SlowDrivers_us"
    free_time_us = "FreeTime_us"
    skipped_cycles = "SkippedCycles"
    hour = "Hour"
    minute = "Minute"
    second = "Second"
    day = "Day"
    month = "Month"
    year = "Year"
    week_day = "WeekDay"
    summer_time = "SummerTime"
    minute_in_day = "MinuteInDay"
    minute_in_day_corr = "MinuteInDayCorr"


class ROOM_VARS(StrEnum):
    fan_command = "FanCommand"
    name = "_RoomName"
    eco_mode = "EcoMode"
    holiday = "InHoliday"
    important = "ImportantRoom"
    calendar_number = "CalendarNumber"
    room_mode = "RoomMode"
    corr_time = "CorrTime"
    winter_holiday = "WinterHoliday"
    summer_holiday = "SummerHoliday"
    min_humidity = "MinHumidity"
    max_humidity = "MaxHumidity"
    actual_temperature = "ActualTemperature"
    requested_temperature = "RequestedTemperature"
    failure = "Failure"
    humidity_problem = "HumidityProblem"
    open_window = "OpenWindow"
    open_window_time = "OpenWindowTime"
    open_window_midnight = "OpenWindowMidnight"
    at_home = "AtHome"
    weekend = "WeekEnd"
    key = "Key"
    humidity_low = "HumidityLow"
    humidity_high = "HumidityHigh"
    heating = "Heating"
    cooling = "Cooling"
    cooling_enabled = "CoolingEnabled"
    heating_enabled = "HeatingEnabled"
    cooling_mode = "CoolingMode"
    manual_correction = "ManualCorrection"
    light_on = "LightOn"
    at_least_one_up = "AtLeastOneUp"
    at_least_one_down = "AtLeastOneDown"
    humidity = "Humidity"
    first_symbol = "FirstSymbol"
    last_symbol = "LastSymbol"
    heating_type = "HeatingType"
    correction_status = "CorrectionStatus"
    correction_time = "CorrectionTime"
    correction_temperature = "CorrectionTemp"
    calendar_temperature = "CalendarTemperature"


class ROOM_MODES(IntEnum):
    CALENDAR = 0
    OFF = 3


class SUNBLIND_VARS(StrEnum):
    failure = "_Failure"
    out_up_1 = "OutUP_1"
    out_up_2 = "OutUP_2"
    out_dn_1 = "OutDN_1"
    out_dn_2 = "OutDN_2"
    manual_up = "Manual_UP"
    manual_dn = "Manual_DN"
    en = "EN"
    dis = "DIS"
    urgent_up = "Urgent_UP"
    sun_dn = "Sun_DN"
    dn = "DN"
    up = "UP"
    step_dn = "StepDN"
    step_up = "StepUP"
    calendar_en = "CalendarEN"
    move_time = "MoveTime"
    reverse_time = "ReverseTime"
    tilt_time = "TiltTime"
    short_down_time = "ShortDownTime"
    step_time = "StepTime"
    full_time_time = "FullTimeTime"
    name = "Name"
    dead_time = "DeadTime"
    state = "State"
    rotation = "Rotation"
    position = "Position"
    command = "COMMAND"
    state2 = "_state"


class SUNBLIND_COMMANDS(IntEnum):
    UP = 0  # move_time
    DOWN = 1  # move_time
    DOWN_TILT = 2  # move_time + reverse_time + tilt_time
    STOP = 3
    TILT_OPEN = 4  # short_down_tim + reverse_time + tilt_time
    STEP_UP = 5  # step_time
    STEP_DOWN = 6  # step_time
    UP_AND_DISABLE = 7  # move_time
    ENABLE = 8
    TILT_OPEN_SHORT = 9  # tilt_time


SUNBLIND_TILT_OPEN = 0
SUNBLIND_TILT_CLOSED = 180
SUNBLIND_RETRACTED = 0
SUNBLIND_EXTENDED = 1000
